@@ -120,15 +120,15 @@ class OldDevelopmentBranchesVisitor(BaseVisitor):
     def visit(self, repo: Repository, **kwargs) -> Section:
         section = Section(rule=self.rule, title='Checked if repo contains abandoned feature branches')
         deadline = datetime.now() - timedelta(days=kwargs['max_days_features'])
-        merged_branches = repo.raw_query(lambda git: git.branch('-r', '--merged', repo.develop.name))
+        merged_branches = repo.raw_query(lambda git: git.branch('-r', '--merged', self.gitflow.develop))
 
         def _check_for_issues(branches: IterableList, name: str):
             for branch in branches:
-                if deadline > branch.commit.authored_datetime.replace(tzinfo=None) \
+                if deadline > branch.commit.committed_datetime.replace(tzinfo=None) \
                         and branch.name not in merged_branches:
                     section.append(Issue.error(
                         '{} {} has not been touched since {}'.format(name, branch.name,
-                                                                     branch.commit.authored_datetime)))
+                                                                     branch.commit.committed_datetime)))
 
         _check_for_issues(branches=repo.branches(folder=self.gitflow.features), name='Feature')
         _check_for_issues(branches=repo.branches(folder=self.gitflow.fixes), name='Fix')
@@ -188,7 +188,7 @@ class MainCommitsAreTaggedVisitor(BaseVisitor):
 
     def visit(self, repo: Repository, **kwargs) -> Section:
         section = Section(rule=self.rule, title='Checked if main repo branch has tagged commits')
-        main_branch = '{}/{}'.format(repo.remote.name, self.gitflow.master)
+        main_branch = '/'.join([repo.remote.name, self.gitflow.master])
         main_commits = repo.raw_query(
             lambda git: git.log(main_branch, '--merges', '--format=format:%H', '--first-parent'),
             predicate=lambda sha: sha)
@@ -208,6 +208,53 @@ class MainCommitsAreTaggedVisitor(BaseVisitor):
         return section
 
 
+class NoDirectCommitsToProtectedBranches(BaseVisitor):
+    """
+    the purposes behind develop and master are different but there is an assumption that at least those two are protected
+
+    the rule is here to check if it is really the case and both branches does not contain direct commits (commits
+    that were pushed directly)
+    """
+
+    @property
+    def rule(self) -> str:
+        return 'no_direct_commits_to_protected_branches'
+
+    def visit(self, repo: Repository, *args, **kwargs) -> Section:
+        section = Section(rule=self.rule, title='Checked if {} and {} contain only merges without direct commits'
+                          .format(self.gitflow.develop, self.gitflow.master))
+        initial_commit = next(iter(repo.raw_query(lambda git: git.log('--format=format:%s', '--reverse'))))
+
+        def _get_direct_commits(branch: str) -> List[str]:
+            merges = repo.raw_query(
+                lambda git: git.log(branch, '--merges', '--format=format:%H', '--first-parent'),
+                predicate=lambda sha: sha)
+            potential_fast_forwards = repo.raw_query(
+                lambda git: git.reflog('show', branch, '--format=format:%H'),
+                predicate=lambda sha: sha)
+            all_commits = repo.raw_query(
+                lambda git: git.log(branch, '--format=format:%H', '--first-parent'),
+                predicate=lambda sha: sha)
+            return [sha for sha in all_commits if sha not in merges and sha not in potential_fast_forwards]
+
+        def _get_issues(direct_commits: List[str], branch: str) -> List[Issue]:
+            issuers = [repo.commit(sha, branch) for sha in direct_commits]
+            issue_msg_fmt = 'Branch {} contains commit "{}" that was pushed directly rather than merged'
+            return [
+                Issue.error(issue_msg_fmt.format(branch, ' '.join([str(commit.hexsha)[:8], commit.summary.strip()])))
+                for commit in issuers if commit and commit.message.lower().strip() != initial_commit.lower().strip()
+            ]
+        
+        dev_branch = '/'.join([repo.remote.name, self.gitflow.develop])
+        main_branch = '/'.join([repo.remote.name, self.gitflow.master])
+
+        develop_results = _get_direct_commits(dev_branch)
+        master_results = _get_direct_commits(main_branch)
+        section.extend(_get_issues(direct_commits=develop_results, branch=dev_branch))
+        section.extend(_get_issues(direct_commits=master_results, branch=main_branch))
+        return section
+
+
 class VersionNamesConventionVisitor(BaseVisitor):
     __doc__ = """checks if release branches and tags follow version naming convention
     
@@ -240,6 +287,45 @@ class VersionNamesConventionVisitor(BaseVisitor):
         return section
 
 
+class DevBranchNamesFollowConvention(BaseVisitor):
+    """
+    sometimes you may wish to have feature and bugfix branch names containing eg. ticket numbers
+
+    the given convention is checked by providing ``name_regex`` as an argument
+
+    if you want to provide different conventions for features and bugfixes, use ``feature_name_regex`` and ``bugfix_name_regex`` respectively
+    """
+
+    @property
+    def rule(self) -> str:
+        return 'dev_branch_names_follow_convention'
+
+    def visit(self, repo: Repository, *args, **kwargs) -> Section:
+        import re
+        features_regex = kwargs.get('name_regex', None) if not kwargs.get('feature_name_regex', None) else kwargs.get(
+            'feature_name_regex')
+        bugfixes_regex = kwargs.get('name_regex', None) if not kwargs.get('bugfix_name_regex', None) else kwargs.get(
+            'bugfix_name_regex')
+        if not features_regex or not bugfixes_regex:
+            raise Exception('Configuration of the rule is invalid: desired convention is not provided')
+
+        feature_prefix = '/'.join([repo.remote.name, self.gitflow.features, ''])
+        bugfix_prefix = '/'.join([repo.remote.name, self.gitflow.fixes, ''])
+
+        def _is_valid(branch: str, regex: str) -> bool:
+            return re.search(regex, branch) is not None
+
+        feature_issuers = [feature.name for feature in repo.branches(self.gitflow.features) if
+                           not _is_valid(branch=feature.name.replace(feature_prefix, ''), regex=features_regex)]
+        bugfix_issuers = [bugfix.name for bugfix in repo.branches(self.gitflow.fixes) if
+                          not _is_valid(branch=bugfix.name.replace(bugfix_prefix, ''), regex=bugfixes_regex)]
+
+        issue_msg_fmt = '{branch} branch does not follow given convention'
+        issues = [Issue.error(issue_msg_fmt.format(branch=branch)) for branch in (feature_issuers + bugfix_issuers)]
+
+        return Section(rule=self.rule, title='Checked if feature and bugfix names follow convention', issues=issues)
+
+
 class DeadReleasesVisitor(BaseVisitor):
     __doc__ = """release branches that are not closed may create a mess in the repository and breaks the master/main 
     branch - releases must be closed as soon as they are deployed to production environment (or just before, 
@@ -257,21 +343,41 @@ class DeadReleasesVisitor(BaseVisitor):
     def visit(self, repo: Repository, *args, **kwargs) -> Section:
         section = Section(rule=self.rule, title='Checked if repo contains abandoned and not removed releases')
         deadline = datetime.now() - timedelta(days=kwargs['deadline_to_close_release'])
-        main_branch = '{}/{}'.format(repo.remote.name, self.gitflow.master)
-        release_branch = '{}/{}/'.format(repo.remote.name, self.gitflow.releases)
-        hotfix_branch = '{}/{}/'.format(repo.remote.name, self.gitflow.hotfixes)
+        main_branch = '/'.join([repo.remote.name, self.gitflow.master])
+        dev_branch = '/'.join([repo.remote.name, self.gitflow.develop])
+        release_branch = '/'.join([repo.remote.name, self.gitflow.releases, ''])
+        hotfix_branch = '/'.format([repo.remote.name, self.gitflow.hotfixes, ''])
 
-        potential_dead_releases = repo.raw_query(lambda git: git.branch('-r', '--no-merged', main_branch),
+        releases_not_merged_to_main = repo.raw_query(lambda git: git.branch('-r', '--no-merged', main_branch),
+                                                 predicate=lambda release: release.strip().startswith(
+                                                     release_branch) or release.strip().startswith(
+                                                     hotfix_branch),
+                                                 map_line=lambda release: repo.branch(release))
+        releases_not_merged_to_develop = repo.raw_query(lambda git: git.branch('-r', '--no-merged', dev_branch),
                                                  predicate=lambda release: release.strip().startswith(
                                                      release_branch) or release.strip().startswith(
                                                      hotfix_branch),
                                                  map_line=lambda release: repo.branch(release))
-        dead_releases = [dead_release for dead_release in potential_dead_releases if
-                         deadline > dead_release.commit.authored_datetime.replace(tzinfo=None)]
 
-        section.extend([Issue.error(
-            '{release} seems abandoned - it has never been merged into the master branch'.format(release=r.name)) for r in
-            dead_releases])
+        potential_dead_releases = [release for release in releases_not_merged_to_main if release in releases_not_merged_to_develop]
+        
+        releases_merged_only_into_develop = [release for release in releases_not_merged_to_main if release not in releases_not_merged_to_develop]
+
+        dead_releases = [dead_release for dead_release in potential_dead_releases if
+                         deadline > dead_release.commit.committed_datetime.replace(tzinfo=None)]
+
+        suspicious_releases = [suspicious_release for suspicious_release in releases_merged_only_into_develop 
+                            if deadline > suspicious_release.commit.committed_datetime.replace(tzinfo=None)]
+
+        section.extend([
+            Issue.error('{release} seems abandoned - it has never been merged into the {master} branch'.format(release=r.name, master=self.gitflow.master)) 
+            for r in dead_releases
+        ])
+
+        section.extend([
+            Issue.warning(description='{release} has been merged back into {develop} but never into {master}'.format(release=r.name, develop=self.gitflow.develop, master=self.gitflow.master))
+            for r in suspicious_releases
+        ])
 
         return section
 
@@ -290,29 +396,53 @@ class DependantFeaturesVisitor(BaseVisitor):
     @arguments_checker(['max_dependant_branches'])
     def visit(self, repo: Repository, *args, **kwargs) -> Section:
         section = Section(rule=self.rule, title='Checked if repo contains dependant feature branches')
-        dev_branch = '{}/{}'.format(repo.remote.name, self.gitflow.develop)
+        dev_branch = '/'.join([repo.remote.name, self.gitflow.develop])
         max_dependant_branches = int(kwargs['max_dependant_branches'])
-        merged_branches = repo.raw_query(lambda git: git.branch('-r', '--merged', repo.develop.name))
-        not_merged = [repo.branch(b.name) for b in repo.branches(self.gitflow.features) if
-                      b.name not in merged_branches]
-        branch_issue_format = '{} seems to depend on other feature branches. It contains following merges: ' + os.linesep + '{}'
+        merged_branches = repo.raw_query(lambda git: git.branch('-r', '--merged', self.gitflow.develop))
+        not_merged = [repo.branch(b.name) for b in repo.branches(self.gitflow.features) if b.name not in merged_branches] + [repo.branch(b.name) for b in repo.branches(self.gitflow.fixes) if b.name not in merged_branches]
+        branch_issue_format = '{} seems to depend on other feature branches. It contains following merges: {}'
+        problematic_branch_sep = os.linesep + '\t\t\t- '
 
         for feature in not_merged:
             name = feature.name
-            merge_commits_query = repo.raw_query(lambda git: git.log('{}..{}'.format(dev_branch, name), '--merges',
+            merge_commits_query = repo.raw_query(lambda git: git.log('..'.join([dev_branch, name]), '--merges',
                                                                      '--first-parent', '--format=format:%H'))
             merge_commits_sha = [commit_sha.strip() for commit_sha in merge_commits_query]
             merge_commits_in_feature = [commit for commit in repo.repo.iter_commits(name, max_count=200) if
                                         commit.hexsha in merge_commits_sha]
-            branch_issues = [commit for commit in merge_commits_in_feature if self.gitflow.develop not in commit.message]
+            branch_issues = [commit for commit in merge_commits_in_feature if
+                             self.gitflow.develop not in commit.message]
 
             if branch_issues:
                 is_limit_exceeded = len(branch_issues) > max_dependant_branches
                 issue_level = Level.ERROR if is_limit_exceeded else Level.WARNING
                 issues_titles = [next(iter(commit.message.split(os.linesep)), None) for commit in branch_issues]
-                issue_desc = branch_issue_format.format(name, os.linesep.join(issues_titles))
+                issue_desc = branch_issue_format.format(name, problematic_branch_sep + problematic_branch_sep.join(issues_titles))
                 section.append(Issue(level=issue_level, description=issue_desc))
 
+        # chained features or features that share commits
+        commits_in_ft_branches = dict()
+        issues_in_ft_branch = dict()
+        branch_issue_format = '{} seems to depend on other feature branches. It shares commits with following branches: {}'
+        for feature in not_merged:
+            commit_hashes = repo.raw_query(lambda git: git.rev_list('--left-right','..'.join([dev_branch, feature.name])))
+            commits_in_ft_branches[feature.name] = commit_hashes
+
+        for feature in commits_in_ft_branches.keys():
+            other_dependant_ft_branches = [
+                ft for ft in commits_in_ft_branches.keys() 
+                if ft != feature and any(hash_c in commits_in_ft_branches[feature] for hash_c in commits_in_ft_branches[ft])
+            ]
+            if other_dependant_ft_branches:
+                issues_in_ft_branch[feature] = other_dependant_ft_branches
+        
+        for feature in issues_in_ft_branch.keys():
+            is_limit_exceeded = len(issues_in_ft_branch[feature]) > max_dependant_branches
+            issue_level = Level.ERROR if is_limit_exceeded else Level.WARNING
+            dependant_branches = problematic_branch_sep + problematic_branch_sep.join(issues_in_ft_branch[feature])
+            issue_desc = branch_issue_format.format(feature, dependant_branches)
+            section.append(Issue(level=issue_level, description=issue_desc))
+
         return section
 
 
@@ -322,7 +452,9 @@ def visitors(gitflow: Gitflow) -> List[BaseVisitor]:
         OldDevelopmentBranchesVisitor(gitflow=gitflow),
         NotScopedBranchesVisitor(gitflow=gitflow),
         MainCommitsAreTaggedVisitor(gitflow=gitflow),
+        NoDirectCommitsToProtectedBranches(gitflow=gitflow),
         VersionNamesConventionVisitor(gitflow=gitflow),
+        DevBranchNamesFollowConvention(gitflow=gitflow),
         DeadReleasesVisitor(gitflow=gitflow),
         DependantFeaturesVisitor(gitflow=gitflow),
     ]
